/*
 * This code is part of the GantzGraf graph visualization libary
 * 
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * http://code.deckfour.org/GantzGraf
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which
 * are not licensed under the terms of the GPL, given that they satisfy one
 * or more of the following conditions:
 * 1) Explicit license is granted to the ProM and ProMimport programs for
 *    usage, linking, and derivative work.
 * 2) Carte blance license is granted to all programs developed at
 *    Eindhoven Technical University, The Netherlands, or under the
 *    umbrella of STW Technology Foundation, The Netherlands.
 * For further exemptions not covered by the above conditions, please
 * contact the author of this code.
 * 
 */
package org.processmining.plugins.fuzzymodel.miner.vis;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import org.jgraph.util.Spline2D;

/**
 * This class implements a spline curve, as specified by a
 * list of quadric bezier segments, in float precision.
 * <p>From a given set of control points (four per segment),
 * arbitrary interpolations may be calculated, useful for
 * full, partial, or relative-point drawing.
 * 
 * @author Christian W. Guenther
 * @author Li Jiafei
 *
 */
public class AnimationCurve implements Cloneable {
	
	/** 
	 * The level of bezier-segment subdivision
	 * applied when approximating the spline as
	 * a polyline (set of straight line segments)
	 */
	protected int NUMBER_INTERPOLATE_POINTS = 81;
	
	/**
	 * X and Y coordinates of the control points
	 */
	protected double[] controlX;
	protected double[] controlY;
	/**
	 * X and Y coordinates of the simplified / approximated points
	 */
	protected double[] pointsX;
	protected double[] pointsY;
	/**
	 * Relative percentage of spline magnitude for each
	 * simplified / approximated point
	 */
	protected double[] pointsPercentage;
	/**
	 * Bounding box of the curve
	 */
	protected Rectangle2D bounds;
	
	/**
	 * Creates a new curve from a set of control points (given in alternating
	 * x and y coordinates, in one array of floating point precision numbers)
	 * @param controlPoints
	 */
	public AnimationCurve(Point2D[] controlPoints) {
		initialize(controlPoints);
	}
	
	/**
	 * Initializes the curve, approximating the polyline and measuring segment
	 * percentages.
	 * @param controlPoints
	 */
	protected void initialize(Point2D[] controlPoints) {
		int numPoints = controlPoints.length;
		controlX = new double[numPoints];
		controlY = new double[numPoints];
		for(int i=0; i<numPoints; i++) {
			controlX[i] = controlPoints[i].getX();
			controlY[i] = controlPoints[i].getY();
		}
		double[][] coords = computeSpline(controlPoints);
		this.pointsX = coords[0];
		this.pointsY = coords[1];
		// derive percentages for control points
		analyzeCurve();
	}
	
	/**
	 * Returns the x-coordinates for all 
	 * simplification / approximation points
	 * @return
	 */
	public double[] getPointsX() {
		return pointsX;
	}
	
	/**
	 * Returns the y-coordinates for all
	 * simplification / approximation points
	 * @return
	 */
	public double[] getPointsY() {
		return pointsY;
	}
	
	/** 
	 * Returns the x-coordinates of all
	 * control points
	 * @return
	 */
	public double[] getControlPointsX() {
		return controlX;
	}
	
	/**
	 * Returns the y-coordinates of all
	 * control points
	 * @return
	 */
	public double[] getControlPointsY() {
		return controlY;
	}
	
	/**
	 * Returns an array of 2 floats, the first being x and the
	 * second the y coordinate of a point at relative position
	 * <code>percentage</code>, between start and end of this
	 * curve. The percentage is given within 0 and 1.
	 * @param percentage
	 * @return
	 */
	public double[] getRelativePoint(float percentage) {
		for(int i=0; i<pointsPercentage.length; i++) {
			if(pointsPercentage[i] == percentage) {
				return new double[] { pointsX[i], pointsY[i] };
			} else if(pointsPercentage[i] > percentage) {
				// between i'th and i-1'th point
				double innerPercentage = (percentage - pointsPercentage[i-1]) 
						/ (pointsPercentage[i] - pointsPercentage[i-1]);
				return interpolate(pointsX[i-1], pointsY[i-1], pointsX[i], pointsY[i], innerPercentage);
			}
		}
		// should not occur!
		return null;
	}
	
	/**
	 * Returns a two-dimensional array of floats, containing
	 * the coordinates of a polyline approximating the given
	 * <code>percentage</code> of the curve, from start to end.
	 * The percentage is given within 0 and 1.
	 * @param percentage
	 * @return
	 */
	public double[][] getPartialPath(double percentage) {
		// look for first control point index having a higher percentage
		// than requested
		int greaterIndex = 0;
		for(int i=0; i<pointsPercentage.length; i++) {
			if(pointsPercentage[i] >= percentage) {
				greaterIndex = i;
				break;
			}
		}
		// copy the pre-simplified points of the approximated path, just
		// before the requested percentage
		double[][] result = new double[2][greaterIndex + 1];
		for(int i=0; i<greaterIndex; i++) {
			result[0][i] = pointsX[i];
			result[1][i] = pointsY[i];
		}
		// add relative point between result control points to the set
		if(pointsPercentage[greaterIndex] == percentage) {
			result[0][greaterIndex] = pointsX[greaterIndex];
			result[1][greaterIndex] = pointsY[greaterIndex];
		} else {
			double innerPercentage = (percentage - pointsPercentage[greaterIndex-1]) 
					/ (pointsPercentage[greaterIndex] - pointsPercentage[greaterIndex-1]);
			double[] interpolated = interpolate(pointsX[greaterIndex-1], pointsY[greaterIndex-1], 
					pointsX[greaterIndex], pointsY[greaterIndex], innerPercentage);
			result[0][greaterIndex] = interpolated[0];
			result[1][greaterIndex] = interpolated[1];
		}
		return result;
	}
	
	/**
	 * Returns the bounding box of this curve
	 * @return
	 */
	public Rectangle2D boundingBox() {
		return bounds;
	}
	
	/**
	 * Checks, whether this curve intersects with the given
	 * rectangle.
	 * @param rect
	 * @return
	 */
	/*public boolean intersects(Rectangle2D rect) {
		if(bounds.intersects(rect)) {
			// check along control point segments
			double x1, y1, x2, y2;
			x1 = controlX[0];
			y1 = controlY[0];
			for(int i=1; i<controlX.length; i++) {
				x2 = controlX[i];
				y2 = controlY[i];
				Rectangle2D segmentBounds = createBoundingBox(x1, y1, x2, y2);
				if(rect.intersects(segmentBounds)) {
					return true;
				}
				x1 = x2;
				y1 = y2;
			}
			return false;
		} else {
			// cannot intersect if bounds fails
			return false;
		}
	}*/
	
	/**
	 * Protected method for creating the bounding box
	 * of this curve
	 */
	protected Rectangle2D createBoundingBox(double x1, double y1, double x2, double  y2) {
		double tmp = 0;
		// swap coordinates if necessary (harmonize bounds)
		if(x1 > x2) {
			tmp = x1;
			x1 = x2;
			x2 = tmp;
		} else if(x1 == x2) {
			x2 += 1f;
		}
		if(y1 > y2) {
			tmp = y1;
			y1 = y2;
			y2 = tmp;
		} else if(y1 == y2) {
			y2 += 1f;
		}
		return new Rectangle2D.Double(x1, y1, x2 - x1, y2 - y1);
	}
	
	/**
	 * Calculates the shortest distance of a given point to any point
	 * on the curve.
	 * @param point
	 * @return
	 */
	public double distance(Point2D point) {
		double distance = Double.MAX_VALUE;
		double current;
		for(int i=1; i<pointsX.length; i++) {
			current = distance(pointsX[i-1], pointsY[i-1], 
					pointsX[i], pointsY[i], 
					(float)point.getX(), (float)point.getY());
			if(current < distance) {
				distance = current;
			}
		}
		return distance;
	}
	
	/**
	 * Compares the distance between a line segment and any given point
	 * @param x1 x coordinate of line point A
	 * @param y1 y coordinate of line point A
	 * @param x2 x coordinate of line point B
	 * @param y2 y coordinate of line point B
	 * @param pX x coordinate of reference point
	 * @param pY y coordinate of reference point
	 * @return
	 */
	protected double distance(double x1, double y1, double x2, double y2, double pX, double pY) {
		// compute squared line segment length / magnitude
		double sqLineMag = squaredLineMagnitude(x1, y1, x2, y2);
		// compute u value
		// (as in Paul Bourke's article/s at:
		// http://local.wasp.uwa.edu.au/~pbourke/geometry/pointline/)
		double u = ((pX - x1) * (x2 - x1) + (pY - y1) * (y2 - y1)) / sqLineMag;
		double result;
		if(u <=0 || u > 1) {
			// shortest distance is not between end points; compute
			// shortest distance to end points
			double dist1 = squaredLineMagnitude(pX, pY, x1, y1);
			double dist2 = squaredLineMagnitude(pX, pY, x2, y2);
			if(dist1 < dist2) {
				result = dist1;
			} else {
				result = dist2;
			}
		} else {
			// shortest distance is between end points
			// use the described formula
			double iX = x1 + u * (x2 - x1);
			double iY = y1 + u * (y2 - y1);
			result = squaredLineMagnitude(pX, pY, iX, iY);
		}
		// return square root of result (actual value)
		return Math.sqrt(result);
	}
	
	/**
	 * Retrieves the squared magnitude,
	 * or length, of a line segment
	 * @param x1 x coordinate of line point A
	 * @param y1 y coordinate of line point A
	 * @param x2 x coordinate of line point B
	 * @param y2 y coordinate of line point B
	 * @return
	 */
	protected double squaredLineMagnitude(double x1, double y1, double x2, double y2) {
		return ((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1));
	}
	
	/**
	 * Internal method for analyzing the bounding box
	 * and curve segment percentages of this curve
	 */
	protected void analyzeCurve() {
		// percentage calculation
		float total = 0.0f;
		this.pointsPercentage = new double[pointsX.length];
		pointsPercentage[0] = 0.0f;
		for(int i=1; i<pointsX.length; i++) {
			total += distance(pointsX[i-1], pointsY[i-1], pointsX[i], pointsY[i]);
			pointsPercentage[i] = total;
		}
		for(int i=1; i<pointsX.length; i++) {
			pointsPercentage[i] /= total;
		}
		// bounding box calculation
		double x1 = Float.MAX_VALUE;
		double y1 = Float.MAX_VALUE;
		double x2 = Float.MIN_VALUE;
		double y2 = Float.MIN_VALUE;
		for(int i=0; i<pointsX.length; i++) {
			if(pointsX[i] < x1) { x1 = pointsX[i]; }
			if(pointsX[i] > x2) { x2 = pointsX[i]; }
			if(pointsY[i] < y1) { y1 = pointsY[i]; }
			if(pointsY[i] > y2) { y2 = pointsY[i]; }
		}
		// prevent zero-area bounding box
		double bWidth = x2 - x1;
		if(bWidth <= 0) {
			bWidth = 1f;
		}
		double bHeight = y2 - y1;
		if(bHeight <= 0) {
			bHeight = 1f;
		}
		this.bounds = new Rectangle2D.Double(x1, y1, bWidth, bHeight);
	}
	
	/**
	 * Simplifies a given spline to an approximating polyline
	 * @param controlX control points x coordinates
	 * @param controlY control points y coordinates
	 * @param numPoints number of points
	 * @return
	 */
	protected double[][] computeSpline(Point2D[] controlPoints) {
		
		int n = controlPoints.length;
		//Point2D p0 = controlPoints[0];
		Point2D pe = controlPoints[n - 1];
		//Point2D p1 = controlPoints[1];
		Point2D p2 = controlPoints[n - 2];
		Spline2D spline;
		int numOfSplinePoints;
		if(n>2){ //curve
			numOfSplinePoints = 81; // 1/t +1 ; t=0.0125
		}
		else{
			numOfSplinePoints = 2; // straight line
		}
		double res[][] = new double[2][numOfSplinePoints];
		//according to the JGraph
		if (n > 2) {
			spline = new Spline2D(controlPoints);
			double[] point = spline.getPoint(0.9875);
			// Extrapolate p2 away from the end point, pe, to avoid integer
			// rounding errors becoming too large when creating the line end
			double scaledX = pe.getX() - ((pe.getX() - point[0]) * 128);
			double scaledY = pe.getY() - ((pe.getY() - point[1]) * 128);
			p2.setLocation(scaledX, scaledY);
			int i=0;
			for (double t = 0; t < 1; t += 0.0125) {
				double[] xy = spline.getPoint(t);				
				res[0][i] = xy[0];
				res[1][i] = xy[1];
				i++;
			//	view.sharedPath.lineTo((float) xy[0], (float) xy[1]);
			}	
		}
		else{ //at least 2 points for an edge.  draw straightline
			for(int i=0;i<n;i++){
				res[0][i] = controlPoints[i].getX();
				res[1][i] = controlPoints[i].getY();
			}
		}
		return res;
	}
	
	/**
	 * Performs bezier subdivision of a given spline segment,
	 * specified as 4 control points in x and y dimension.
	 * @param vX x coordinates of the 4 control points
	 * @param vY y coordinates of the 4 control points
	 * @param t
	 * @return
	 */
	/*protected double[] bezier(double[] vX, double[] vY, double t) {
		int degree = 3;	// 4 control points
		double[][] vTempX = new double[degree + 1][degree + 1];
		double[][] vTempY = new double[degree + 1][degree + 1];
		// copy control points
		for(int j=0; j<=degree; j++) {
			vTempX[0][j] = vX[j];
			vTempY[0][j] = vY[j];
		}
		// triangle computation
		for(int i=1; i<=degree; i++) {
			for(int j=0; j<=(degree - i); j++) {
				vTempX[i][j] = (1.0f - t) * (vTempX[i - 1][j]) + t * (vTempX[i - 1][j + 1]);
				vTempY[i][j] = (1.0f - t) * (vTempY[i - 1][j]) + t * (vTempY[i - 1][j + 1]);
			}
		}
		// return result
		double[] result = new double[2];
		result[0] = vTempX[degree][0];
		result[1] = vTempY[degree][0];
		return result;
	}*/
	
	/**
	 * Interpolates linearly between two given points by a relative percentage
	 * @param x1 A, x coordinate
	 * @param y1 A, y coordinate
	 * @param x2 B, x coordinate
	 * @param y2 B, y coordinate
	 * @param percentage percentage of the requested point between A and B.
	 * @return
	 */
	protected double[] interpolate(double x1, double y1, double x2, double y2, double percentage) {
		double[] interpolated = new double[2];
		double counterPercentage = 1.0f - percentage;
		interpolated[0] = (counterPercentage * x1) + (percentage * x2);
		interpolated[1] = (counterPercentage * y1) + (percentage * y2);
		return interpolated;
	}
	
	/**
	 * Calculates the distance between two given points
	 * @param x1 A, x coordinate
	 * @param y1 A, y coordinate
	 * @param x2 B, x coordinate
	 * @param y2 B, y coordinate
	 * @return
	 */
	protected double distance(double x1, double y1, double x2, double y2) {
		// behold the power of Pythagoras..
		double a = x1 - x2;
		double b = y1 - y2;
		return Math.sqrt((a * a) + (b * b));
	}
	
	/**
	 * Clones this curve object
	 */
	public Object clone() {
		AnimationCurve clone = null;
		try {
			clone = (AnimationCurve)super.clone();
		} catch(CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
		clone.controlX = controlX.clone();
		clone.controlY = controlY.clone();
		clone.pointsX = pointsX.clone();
		clone.pointsY = pointsY.clone();
		clone.pointsPercentage = pointsPercentage.clone();
		clone.bounds = (Rectangle2D)bounds.clone();
		return clone;
	}

}
