/*
 * This code is part of the GantzGraf graph visualization libary
 * 
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * http://code.deckfour.org/GantzGraf
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which are not
 * licensed under the terms of the GPL, given that they satisfy one or more of
 * the following conditions: 1) Explicit license is granted to the ProM and
 * ProMimport programs for usage, linking, and derivative work. 2) Carte blance
 * license is granted to all programs developed at Eindhoven Technical
 * University, The Netherlands, or under the umbrella of STW Technology
 * Foundation, The Netherlands. For further exemptions not covered by the above
 * conditions, please contact the author of this code.
 */
package org.processmining.plugins.fuzzymodel.miner.vis;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingConstants;

import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.AttributeMap.SerializablePoint2D;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.fuzzymodel.FMNode;
import org.processmining.models.graphbased.directed.fuzzymodel.MutableFuzzyGraph;
import org.processmining.models.graphbased.directed.jgraph.JGraphVisualizationPanel;
import org.processmining.models.graphbased.directed.jgraph.ProMGraphModel;
import org.processmining.models.graphbased.directed.jgraph.ProMJGraph;
import org.processmining.models.graphbased.directed.jgraph.ProMJGraphVisualizer;
import org.processmining.models.graphbased.directed.jgraph.elements.ProMGraphCell;
import org.processmining.models.graphbased.directed.jgraph.elements.ProMGraphEdge;
import org.processmining.models.graphbased.directed.jgraph.elements.ProMGraphPort;
import org.processmining.models.graphbased.directed.jgraph.views.JGraphEdgeView;
import org.processmining.models.graphbased.directed.jgraph.views.JGraphShapeView;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;

/**
 * This class wraps access to the Graphviz Dot executable, which is used for
 * graph layout. The executable is invoked in a transparent manner, for which
 * objects of this class implement a convenient handle.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * @author Li Jiafei
 */
public class AnimationFMJGraphParser {

	protected MutableFuzzyGraph graph;
	protected ProMJGraph fmJGraph;
	protected JGraphFacade facade;
	protected JGraphVisualizationPanel graphPanel;
	protected PluginContext context;

	/**
	 * Scaling factor
	 */
	//protected double scale;
	/**
	 * Bounding box dimensions (parsed from dot)
	 */
	protected double bBoxWidth;
	protected double bBoxHeight;
	/**
	 * Hash maps for storing graph element coordinates, referenced by their
	 * respective IDs.
	 */
	protected HashMap<String, double[]> nodeCoordinates;
	protected HashMap<String, Point2D[]> edgeCoordinates;

	public AnimationFMJGraphParser(PluginContext context, MutableFuzzyGraph graph) {
		this.context = context;
		this.graph = graph;
		//this.graph.setLayedOut(false);
	//	this.graph.signalViews();		
		this.graphPanel = ProMJGraphVisualizer.visualizeGraph(this.graph);
		this.fmJGraph = graphPanel.getGraph();		
	//	this.fmJGraph = new ProMJGraph(new ProMGraphModel(graph), new ViewSpecificAttributeMap());
		layOutFMJGraph(fmJGraph);
		parse();
	}

	protected void layOutFMJGraph(ProMJGraph jgraph) {

		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();
		layout.setDeterministic(false);
		layout.setCompactLayout(false);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(20);
		layout.setOrientation(graph.getAttributeMap().get(AttributeMap.PREF_ORIENTATION, SwingConstants.SOUTH));

		this.facade = new JGraphFacade(jgraph);
		facade.setOrdered(true);
		facade.setEdgePromotion(true);
		facade.setIgnoresCellsInGroups(false);
		facade.setIgnoresHiddenCells(false);
		facade.setIgnoresUnconnectedCells(false);
		facade.setDirected(false);
		facade.resetControlPoints();

		facade.run(layout, false);

		animFixParallelEdges(facade, 15);

		Map<?, ?> nested = facade.createNestedMap(true, false);

		jgraph.getGraphLayoutCache().edit(nested);
		jgraph.repositionToOrigin();
		jgraph.setUpdateLayout(layout);

	}

	@SuppressWarnings("unchecked")
	private void animFixParallelEdges(JGraphFacade facade, double spacing) {
		ArrayList edges = new ArrayList(facade.getEdges());
		for (Object edge : edges) {
			List<Object> points = facade.getPoints(edge);
			if (points.size() != 2) {
				continue;
			}
			Object sourceCell = facade.getSource(edge);
			Object targetCell = facade.getTarget(edge);
			Object sourcePort = facade.getSourcePort(edge);
			Object targetPort = facade.getTargetPort(edge);
			Object[] between = facade.getEdgesBetween(sourcePort, targetPort, false);
			if ((between.length == 1) && !(sourcePort == targetPort)) {
				continue;
			}
			Rectangle2D sCP = facade.getBounds(sourceCell);
			Rectangle2D tCP = facade.getBounds(targetCell);
			Point2D sPP = GraphConstants.getOffset(((ProMGraphPort) sourcePort).getAttributes());
			// facade. getBounds (sourcePort ) ;

			if (sPP == null) {
				sPP = new Point2D.Double(sCP.getCenterX(), sCP.getCenterY());
			}
			Point2D tPP = GraphConstants.getOffset(((ProMGraphPort) targetPort).getAttributes());
			// facade.getBounds(sourcePort);

			if (tPP == null) {
				tPP = new Point2D.Double(tCP.getCenterX(), tCP.getCenterY());
			}

			if (sourcePort == targetPort) {
				assert (sPP.equals(tPP));
				double x = sPP.getX();
				double y = sPP.getY();
				for (int i = 2; i < between.length + 2; i++) {
					List<Point2D> newPoints = new ArrayList<Point2D>(5);
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y));
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x, y - (2 * spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing), y - (spacing / 2 + i * spacing)));
					facade.setPoints(between[i - 2], newPoints);
				}

				continue;
			}

			double dx = (sPP.getX()) - (tPP.getX());
			double dy = (sPP.getY()) - (tPP.getY());
			double mx = (tPP.getX()) + dx / 2.0;
			double my = (tPP.getY()) + dy / 2.0;
			double slope = Math.sqrt(dx * dx + dy * dy);
			for (int i = 0; i < between.length; i++) {
				List<Point2D> newPoints = new ArrayList<Point2D>(3);
				double pos = 2 * i - (between.length - 1);
				if (facade.getSourcePort(between[i]) == sourcePort) {
					newPoints.add(sPP);
					newPoints.add(tPP);
				} else {
					newPoints.add(tPP);
					newPoints.add(sPP);
				}
				if (pos != 0) {
					pos = pos / 2;
					double x = mx + pos * spacing * dy / slope;
					double y = my - pos * spacing * dx / slope;
					newPoints.add(1, new SerializablePoint2D.Double(x, y));
				}
				facade.setPoints(between[i], newPoints);
			}
		}
	}

	/**
	 * Returns the scale factor of the parsed graph
	 */
	/*
	 * public double getScale() { return scale;
	 * 
	 * }
	 */

	/**
	 * Returns the bounding box width of the parsed graph
	 */
	public double getBoundsWidth() {

		return bBoxWidth;
	}

	/**
	 * Returns the bounding box height of the parsed graph
	 */
	public double getBoundsHeight() {

		return bBoxHeight;
	}

	/**
	 * Returns the set of all node IDs in the parsed graph
	 */
	public Set<String> getNodeIds() {
		return nodeCoordinates.keySet();
	}

	/**
	 * Returns the set of all edge IDs in the parsed graph
	 */
	public Set<String> getEdgeIds() {
		return edgeCoordinates.keySet();
	}

	/**
	 * Returns the node coordinates for a given node ID.
	 * <p>
	 * The coordinates are returned in an array, in order:
	 * <code>{x, y, width, height}</code>.
	 */
	public double[] getNodeCoordinates(String nodeId) {
		return nodeCoordinates.get(nodeId);
	}

	/**
	 * Returns the edge coordinates for a given edge ID.
	 * <p>
	 * The control point coordinates are returned in an array, in order:
	 * <code>{x1, y1, x2, y2,...}</code>. Every four control points make up a
	 * quadric bezier segment of a spline.
	 */
	public Point2D[] getEdgeCoordinates(String edgeId) {
		return edgeCoordinates.get(edgeId);
	}

	/**
	 * Parse attributed plain dot output from the given buffered reader, storing
	 * the results in internal maps and fields.
	 */
	@SuppressWarnings("unchecked")
	protected void parse() {
		// initialize data structures
		//scale = 0;
		bBoxWidth = 0;
		bBoxHeight = 0;
		nodeCoordinates = new HashMap<String, double[]>();
		edgeCoordinates = new HashMap<String, Point2D[]>();
		//	Object[] allComponents = fmJGraph.getComponents();
		ArrayList vertices = new ArrayList(facade.getVertices());

		// matches the graph identifier, parse general
		// attributes

		// bBoxWidth = facade.getCellBounds().getWidth();
		 //bBoxHeight = facade.getCellBounds().getHeight();
		  bBoxWidth = facade.getGraphBounds().getWidth();
		  bBoxHeight = facade.getGraphBounds().getHeight();
		//bBoxWidth = fmJGraph.getBounds().getWidth();
		//bBoxHeight = fmJGraph.getBounds().getHeight();
		// read the information of each component in the ProMJGraph
		for (Object vertex : vertices) {

			if (vertex instanceof ProMGraphCell) {
				ProMGraphCell node = (ProMGraphCell) vertex;
				JGraphShapeView nodeView = node.getView();
				Rectangle2D nodeBound = nodeView.getBounds();

				
				// parse node attributes
				double x = nodeBound.getX();
				double y = nodeBound.getY();
				double width = nodeBound.getWidth();
				double height = nodeBound.getHeight();
				// correct y parameter to set origin to upper left
				// and set all coordinates in [0,1]
		/*		y = (bBoxHeight - y) / bBoxHeight;
				x /= bBoxWidth;
				width /= bBoxWidth;
				height /= bBoxHeight;
				y -= (height / 2);
				x -= (width / 2);*/
				nodeCoordinates.put(((FMNode) node.getNode()).id(), new double[] { x, y, width, height });
			}
		}

		ArrayList edges = new ArrayList(facade.getEdges());
		for (Object edge : edges) {
			ProMGraphEdge graphEdge = (ProMGraphEdge) edge;
			JGraphEdgeView edgeView = graphEdge.getView();
			int size = edgeView.getPointCount();
			// parse control points
			Point2D[] ctrlPoints = new Point2D[size];

			for (int i = 0; i < size; i++) {
				Point2D controlPoint = edgeView.getPoint(i);
				if (controlPoint != null) {
				/*	double x = controlPoint.getX() / bBoxWidth;
					double y = (bBoxHeight - controlPoint.getY()) / bBoxHeight;*/
					double x = controlPoint.getX();
					double y = controlPoint.getY();
					ctrlPoints[i] = new Point2D.Double(x, y);
				}
			}

			ProMGraphCell sourceCell = (ProMGraphCell) (facade.getSource(edge));
			ProMGraphCell targetCell = (ProMGraphCell) facade.getTarget(edge);
			String source = ((FMNode) sourceCell.getNode()).id();
			String target = ((FMNode) targetCell.getNode()).id();
			edgeCoordinates.put(source + "->" + target, ctrlPoints);
		}
	}

	public String toString() {
		StringBuilder buffer = new StringBuilder();
		buffer.append(bBoxWidth + "x" + bBoxHeight + "\n");
		for (String key : nodeCoordinates.keySet()) {
			buffer.append("node " + key + ":");
			for (double c : nodeCoordinates.get(key)) {
				buffer.append(" " + c);
			}
			buffer.append("\n");
		}
		for (String key : edgeCoordinates.keySet()) {
			buffer.append("edge " + key + ":");
			for (Point2D c : edgeCoordinates.get(key)) {
				buffer.append(" " + c);
			}
			buffer.append("\n");
		}
		return buffer.toString();
	}

}
