/*
 * Copyright (c) 2007 Christian W. Guenther (christian@deckfour.org)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * EXEMPTION:
 * 
 * License to link and use is also granted to open source programs which
 * are not licensed under the terms of the GPL, given that they satisfy one
 * or more of the following conditions:
 * 1) Explicit license is granted to the ProM and ProMimport programs for
 *    usage, linking, and derivative work.
 * 2) Carte blance license is granted to all programs developed at
 *    Eindhoven Technical University, The Netherlands, or under the
 *    umbrella of STW Technology Foundation, The Netherlands.
 * For further exemptions not covered by the above conditions, please
 * contact the author of this code.
 * 
 */
package org.processmining.plugins.fuzzymodel.miner.vis.anim;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.graphbased.directed.fuzzymodel.FMClusterNode;
import org.processmining.models.graphbased.directed.fuzzymodel.FMNode;
import org.processmining.models.graphbased.directed.fuzzymodel.MutableFuzzyGraph;
import org.processmining.models.graphbased.directed.fuzzymodel.impl.FMEdgeImpl;
import org.processmining.models.graphbased.directed.fuzzymodel.util.FuzzyMinerLog;
import org.processmining.plugins.fuzzymodel.miner.vis.AnimationCurve;
import org.processmining.plugins.fuzzymodel.miner.vis.AnimationFMJGraphParser;
import org.processmining.plugins.fuzzymodel.miner.vis.anim.TaskAnimation.Type;
import org.processmining.plugins.fuzzymodel.miner.vis.ui.FMProgressPanel;

/**
 * @author Christian W. Guenther (christian@deckfour.org)
 *@author Jiafei Li (jiafei@jlu.edu.cn)
 */
public class Animation {
	
	protected XLog log;
	protected MutableFuzzyGraph graph;
	protected Set<FMNode> graphNodes;
	protected Set<FMEdgeImpl> graphEdges;
	protected AnimationFMJGraphParser parser;
	
	protected int maxLookahead = 5;
	protected int maxExtraLookahead = 2;
	
	protected HashMap<String,TaskAnimation> taskAnimations;
	protected HashMap<String,ArcAnimation> arcAnimations;
	protected ArrayList<TaskAnimation> primitiveTaskAnimationList;
	protected ArrayList<TaskAnimation> clusterTaskAnimationList;
	protected ArrayList<ArcAnimation> arcAnimationList;
	protected ArcAnimation caseAnimation;
	
	protected long startTime;
	protected long endTime;
	
	protected int minTraverseCount;
	protected int maxTraverseCount;
	protected int bestQuarterTraverseCount;
	protected long meanCaseDuration;
	
	protected long fadeTime;
	protected long meanBetweenTime;
	protected PluginContext context;
	

	
	public Animation(PluginContext context,MutableFuzzyGraph graph, XLog log, int maxLookahead, int maxExtraLookahead) {
		this.context = context;
		this.log = log;
		this.graph = graph;
		this.maxLookahead = maxLookahead;
		this.maxExtraLookahead = maxExtraLookahead;
		this.graphNodes = graph.getNodes();
		this.graphEdges = graph.getEdgeImpls();
		this.taskAnimations = new HashMap<String,TaskAnimation>();
		this.arcAnimations = new HashMap<String,ArcAnimation>();
		this.primitiveTaskAnimationList = new ArrayList<TaskAnimation>();
		this.clusterTaskAnimationList = new ArrayList<TaskAnimation>();
		this.arcAnimationList = new ArrayList<ArcAnimation>();
		this.startTime = Long.MAX_VALUE;
		this.endTime = Long.MIN_VALUE;
	}
	
	/**
	 * @throws IOException  
	 */
	public static Animation generate(PluginContext context,MutableFuzzyGraph graph, XLog log, int maxLookahead, int maxExtraLookahead,FMProgressPanel progress) 
			throws IndexOutOfBoundsException, IOException {
//		Progress progress = context.getProgress();
		
		progress.changeProgress(-1);
		progress.changeProgressCaption("Projecting log onto graph...");
		//log.getLogSummary();	// trigger filtering
	//	XLogInfo logInfo = XLogInfoFactory.createLogInfo(log);
		
		Animation animation = new Animation(context,graph, log, maxLookahead, maxExtraLookahead);
		animation.initialize(progress);
		progress.changeProgress(-1);
		progress.changeProgressCaption("Finished.");
		return animation;
	}
	
	public AnimationFMJGraphParser getParser() {
		return parser;
	}
	
	public List<TaskAnimation> getPrimitiveTaskAnimations() {
		return primitiveTaskAnimationList;
	}
	
	public List<TaskAnimation> getClusterTaskAnimations() {
		return clusterTaskAnimationList;
	}
	
	public List<ArcAnimation> getArcAnimations() {
		return arcAnimationList;
	}
	
	public ArcAnimation getCaseAnimation() {
		return caseAnimation;
	}
	
	public long getStart() {
		return startTime;
	}
	
	public long getEnd() {
		return endTime;
	}
	
	public int getMinTraverseCount() {
		return minTraverseCount;
	}
	
	public int getMaxTraverseCount() {
		return maxTraverseCount;
	}
	
	public int getBestQuarterTraverseCount() {
		return bestQuarterTraverseCount;
	}
	
	public long getFadeTime() {
		return fadeTime;
	}
	
	public long getMeanBetweenTime() {
		return meanBetweenTime;
	}
	
	public long getMeanCaseDuration() {
		return meanCaseDuration;
	}
	
	public int getActivityBetween(long left, long right) {
		int activity = 0;
		for(TaskAnimation anim : taskAnimations.values()) {
			activity += anim.getActivityBetween(left, right);
		}
		return activity;
	}
	
	public void initialize(FMProgressPanel progress) throws IndexOutOfBoundsException {
		//int progressMax = log.getLogSummary().getNumberOfAuditTrailEntries();
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log);
		int progressMax = logInfo.getNumberOfEvents();
		progress.changeProgress(-1);
		progress.changeProgressCaption("Creating and parsing graph layout...");
		// retrieve coordinates of elements
	/*	StringWriter writer = new StringWriter();
		try {
			graph.writeToDot(writer);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}*/
		parser = new AnimationFMJGraphParser(context,graph);
		
		// initialize data structures
		progress.changeProgressCaption("Initializing animation...");
		// initialize node / task animations from graph
		for(FMNode node : graphNodes) {
			String id = node.id();
			Type type = TaskAnimation.Type.PRIMITIVE;
			if(node instanceof FMClusterNode) {
				type = TaskAnimation.Type.CLUSTER;
			}
			double[] coords = parser.getNodeCoordinates(id);
			TaskAnimation anim = new TaskAnimation(node, id, type, 
					coords[0], coords[1], coords[2], coords[3]);
			taskAnimations.put(id, anim);
			if(type.equals(TaskAnimation.Type.CLUSTER)) {
				clusterTaskAnimationList.add(anim);
			} else {
				primitiveTaskAnimationList.add(anim);
			}
		}
		// initialize edge / arc animations from graph
		for(FMEdgeImpl edge : graphEdges) {
			String sourceId = edge.getSource().id();
			String targetId = edge.getTarget().id();
			String id = createEdgeId(edge.getSource(), edge.getTarget());
			Point2D[] coords = parser.getEdgeCoordinates(id);
			AnimationCurve curve = new AnimationCurve(coords);
			ArcAnimation anim = new ArcAnimation(id, sourceId, targetId, curve);
			arcAnimations.put(id, anim);
			arcAnimationList.add(anim);
		}
		// initialize case animation
		this.caseAnimation = new ArcAnimation("cases", "start", "end", null);
		// prepare for parsing
		//progress.setMinMax(0, progressMax);
		progress.setMinimum(0);
		progress.setMaximum(progressMax);
		// parse all instances from log
		for(int i=0; i<logInfo.getNumberOfTraces(); i++) {
			progress.changeProgressCaption("Reading log: trace " + (i + 1 ) + " / " + logInfo.getNumberOfTraces());
			parse(log.get(i), progress);
		}
		// prepare for finalizing
		progress.changeProgress(-1);
		progress.changeProgressCaption("Preparing playback...");
		// finalize all data structures
		for(ArcAnimation anim : arcAnimations.values()) {
			anim.finalizeData();
		}
		for(TaskAnimation anim : taskAnimations.values()) {
			anim.finalizeData();
		}
		this.caseAnimation.finalizeData();
		// calculate traverse counts
		int[] traverseCounts = new int[arcAnimationList.size()];
		for(int i=0; i<traverseCounts.length; i++) {
			traverseCounts[i] = arcAnimationList.get(i).getTraverseCount();
		}
		if(traverseCounts.length > 0) {
			Arrays.sort(traverseCounts);
			this.minTraverseCount = traverseCounts[0];
			this.maxTraverseCount = traverseCounts[traverseCounts.length - 1];
			int bqtcIndex = (int)(traverseCounts.length * 0.75f);
			this.bestQuarterTraverseCount = traverseCounts[bqtcIndex];
		} else {
			this.minTraverseCount = 0;
			this.maxTraverseCount = 0;
			this.bestQuarterTraverseCount = 0;
		}
		// calculate fade time
		meanBetweenTime = 0;
		meanCaseDuration = 0;
		long start, end, tmp;
		int divisor = 0;
		XTrace ateList;
		for(int i=0; i<logInfo.getNumberOfTraces(); i++) {
			ateList = log.get(i);
			/*if(ateList.size() < 2) {
				continue;
			}*/
			divisor++;
		/*	start = ateList.get(0).getTimestamp().getTime();
			end = ateList.get(ateList.size() - 1).getTimestamp().getTime();*/
			start = FuzzyMinerLog.getEventTime(ateList.get(0)).getTime();
			end = FuzzyMinerLog.getEventTime(ateList.get(ateList.size() - 1)).getTime();
     		
			if(ateList.size()>=2)
     		{
			   tmp = (end - start) / (ateList.size() - 1);
			   meanCaseDuration += (end - start);
     		}
     		else
     		{
     			tmp = 1000;
     			meanCaseDuration += tmp;     			
     		}
			meanBetweenTime += tmp;
			
			//meanCaseDuration += (end - start);
		}
		meanCaseDuration /= logInfo.getNumberOfTraces();
		meanBetweenTime /= divisor;
		this.fadeTime = meanBetweenTime * 5;
		this.startTime -= 1000;
		this.endTime += this.fadeTime;
	}
	
	protected void parse(XTrace instance, FMProgressPanel progress) throws IndexOutOfBoundsException {
		int counter = 0;
		//String caseId = instance.getName();
		String caseId = ((XAttributeLiteralImpl)instance.getAttributes().get("concept:name")).getValue();
		XTrace ateList = instance;
		// insert case animation
		XEvent startEvt,endEvt;
		long start,end;
		startEvt =ateList.get(0);
		endEvt = ateList.get(ateList.size() - 1);
		start =  FuzzyMinerLog.getEventTime(startEvt).getTime();
		end = FuzzyMinerLog.getEventTime(endEvt).getTime();
		TokenAnimation caseTokenAnim = new TokenAnimation(caseId, start,end);
		this.caseAnimation.addTokenAnimation(caseTokenAnim);
		// parse event list
		boolean[] seeded = new boolean[ateList.size()];
		Arrays.fill(seeded, false);
		XEvent ref, comp;
		FMNode refNode, compNode;
		boolean swallowed;
		// step through event list
		for(int i=0; i<ateList.size(); i++) {
			counter++;
			if(counter > 299) {
				progress.changeProgress(progress.getValue() + counter);
				counter = 0;
			}
			swallowed = true;	// reset swallowed marker
			ref = ateList.get(i);
			refNode = resolveNode(ref);
			if(refNode == null) {
				continue;	// invalid event, ignore
			}
			// look ahead for potentially connected events
			int countdown = maxExtraLookahead;
			for(int j=i+1; (j<=i+maxLookahead) && (j<ateList.size()); j++) {
				comp = ateList.get(j);
				compNode = resolveNode(comp);
				if(compNode == null) {
					continue;	// invalid event, ignore
				}
			//	FMEdge edge = graphEdges.getEdge(refNode, compNode);
				FMEdgeImpl edge = null;
				for(FMEdgeImpl fmEdge: graphEdges){
					if(fmEdge.getSource().equals(refNode)&&fmEdge.getTarget().equals(compNode)){
						edge = fmEdge;
					}
				}
					
				if(edge != null) {
					// connected, mark target position
					seeded[j] = true;
					// mark as non-swallowed
					swallowed = false;
					// create and add token animation
					/*TokenAnimation tokenAnim = new TokenAnimation(caseId, 
							ref.getTimestamp().getTime(), comp.getTimestamp().getTime());*/
					TokenAnimation tokenAnim = new TokenAnimation(caseId, 
							FuzzyMinerLog.getEventTime(ref).getTime(),
							FuzzyMinerLog.getEventTime(comp).getTime());
					ArcAnimation arcAnim = arcAnimations.get(createEdgeId(refNode, compNode));
					arcAnim.addTokenAnimation(tokenAnim);
				}
				// stop looking after the extra lookahead after
				// finding the first valid connection.
				if(swallowed == false) {
					countdown--;
					if(countdown < 0) {
						break;
					}
				}
			}
			// create and add task animation
			TaskAnimationKeyframe taskKeyframe = new TaskAnimationKeyframe(caseId, 
					FuzzyMinerLog.getEventTime(ref).getTime(), !seeded[i], swallowed);
			TaskAnimation taskAnim = taskAnimations.get(refNode.id());
			taskAnim.addKeyframe(taskKeyframe);
			// update boundaries
			updateBoundaries(FuzzyMinerLog.getEventTime(ref).getTime());
		}
		progress.changeProgress(progress.getValue() + counter);
	}
	
	protected FMNode resolveNode(XEvent ate) {
		//String element = ate.getElement();
		String element = FuzzyMinerLog.getEventName(ate);
		if(element.startsWith("Cluster")) {
			// referred node is a cluster
			for(FMClusterNode cluster : graph.getClusterNodes()) {
				if(cluster.getElementName().equals(element)) {
					return cluster;
				}
			}
		}
		// look for primitive node
		//int index = graph.getEventClassIndex(ate.getElement(), ate.getType());
		int index = graph.getEventClassIndex(FuzzyMinerLog.getEventName(ate), FuzzyMinerLog.getEventType(ate));
		if(index >= 0) {
			return graph.getNodeMappedTo(index);
		} else {
			return null;
		}
	}
	
	protected String createEdgeId(FMNode source, FMNode target) {
		return source.id() + "->" + target.id();
	}
	
	protected void updateBoundaries(long time) {
		if(time < startTime) {
			startTime = time;
		}
		if(time > endTime) {
			endTime = time;
		}
	}
	

	
}
